/* TODO Code in this script duplicated in popup and event page,
 * better to move all of this functions to the event page and call from popup using messages?
 */

// eslint-disable-next-line no-unused-vars
const stageUrl = 'https://mjqj47yt17.execute-api.us-east-1.amazonaws.com/test/';
const stageUrlNew = 'https://api.intelligent-speaker.com/';
const clientId = 'k44r81bdug8tm4721kbfuqbm5';
const clientIdOld = '722jaa8aarrubh9lc1v3ff2ibd';

// eslint-disable-next-line no-unused-vars
const REDIRECT_URI = 'https://api.intelligent-speaker.com/cognito-callback';

const uninstallURL =
    'https://docs.google.com/forms/d/e/' +
    '1FAIpQLSdDNyd3COJzZEw9sHXE34Y2Ercrq0klL1hCn0Or2mZapG7UQw/viewform' +
    '?usp=pp_url&entry.1897026270&entry.523383874=';

const CHARS_IN_MINUTE = 1040;  // eslint-disable-line no-unused-vars

// eslint-disable-next-line no-unused-vars
function _getRandomString() {
    return String(crypto.getRandomValues(new Uint32Array(3))).replace(/,/g, '');
}


/*
 * TODO read lang from store only - persist that value on install/update?
 */
function _getLang() {  // eslint-disable-line no-unused-vars
    if (navigator.language === 'pt') {
        return 'pt_PT';
    }
    return navigator.language.slice(0, 2);
}


async function _isLangSupported(lang) {  // eslint-disable-line no-unused-vars
    // eslint-disable-next-line no-undef
    return (location.pathname === '/eventPage.html' ? await _getPopupDOM() : document)
        .querySelectorAll(`[data-lang=${lang}]`);
}

function _getMessages(lang, isFallback) {  // eslint-disable-line no-unused-vars
    return fetch(chrome.runtime.getURL(`_locales/${lang}/messages.json`))
        .then(resp => resp.json())
        .then(function(resp) {
            // function instead of `=>` against interpreting `{` as function body opening (error)
            return {'messages': resp, isFallback};
        })
        .catch(_ => _getMessages('en', true));
    // Example of unsupported locale: 'eo' (Esperanto) - emailed by automatic report system
}


function _getActiveTab(callback) {  // eslint-disable-line no-unused-vars
    // https://developer.chrome.com/extensions/tabs#method-query
    chrome.tabs.query({'active': true, 'currentWindow': true}, tabs => {
        find: for (const tab of tabs) {  // Label used just to practice with it
            if (tab.url) {
                // Opened devtools do not has url.
                // Not only during development but user can also has an opened devtools.
                // Tested in Google Chrome.
                callback(tab);
                break find;
            }
            console.error(
                'Not found active tabs (click on tab; only when devtools is opened)'
            );
        }

    });
}


/**
 * This function is the decorator for any calls where id_token is used.
 *
 * id_token is valid for one hour.
 *
 * @see {@link http://docs.aws.amazon.com/cognito/latest/developerguide/token-endpoint.html}
 * @see {@https://stackoverflow.com/questions/46491811}
 */
async function _keepTokenFresh(storage, client_id=clientId) {  // eslint-disable-line no-unused-vars
    console.log('_keepTokenFresh, storage:', storage);
    if (!storage) {
        storage = await _chromeStorageSyncGet();
    }
    if (_getValueFromIdToken('exp', storage) < (Date.now() / 1000 | 0)) {

        try {
            const resp = await fetch(
                'https://intelligentspeaker.auth.us-east-1.amazoncognito.com/oauth2/token', {
                    'method': 'POST',
                    'headers': {
                        'Content-Type': 'application/x-www-form-urlencoded'
                    },
                    'body': 'grant_type=refresh_token' +
                            '&client_id=' + client_id +
                            '&refresh_token=' + storage['refresh_token']
                });
            const json = await resp.json();
            if (json['error'] === 'invalid_grant') {
                return _keepTokenFresh(storage, clientIdOld);
            }
            const jsonTokenNew = {'id_token': json['id_token']};

            console.debug('New id token:', jsonTokenNew['id_token']);

            chrome.storage.sync.set(jsonTokenNew);

            return {...storage, ...jsonTokenNew};  // Merge
        } catch(e) {
            console.log(
                'TODO maybe normal case when no internet connection - handle this without errors');
            alert('Intelligent Speaker: error, try later');  // TODO i18n
            stopPlaying();  // eslint-disable-line no-undef
        }
    } else {
        return storage;
    }
}

function _chromeStorageSyncGet() {
    return new Promise(resolve => chrome.storage.sync.get(resp => resolve(resp)));
}

function openCognitoPage(state) {  // eslint-disable-line no-unused-vars

    // 'response_type' is 'code' but not 'token' because backend need to distinguish clients
    // https://developer.chrome.com/extensions/tabs#method-create
    // http://docs.aws.amazon.com/cognito/latest/developerguide/cognito-user-pools-ux.html
    chrome.tabs.create({'url': 'https://auth.intelligent-speaker.com/signup' +
         '?response_type=code' +
         '&client_id=' + clientId +  // eslint-disable-line no-undef
         '&redirect_uri=' + REDIRECT_URI + // eslint-disable-line no-undef

        // eslint-disable-next-line no-undef
        '&state=' + state
    });
}
async function _init(state) {  // eslint-disable-line no-unused-vars
    let initData = {};

    try {
        const resp = await fetch(`${stageUrlNew}init`, {
            'method': 'POST',
            'body': state
        });
        initData = await resp.json();
        console.log('Init data:', initData);
    } catch(e) {
        alert('Intelligent Speaker: error, try later');  // TODO i18n
        stopPlaying();  // eslint-disable-line no-undef
        throw e;
    }

    /** TODO about 'isCognitoWasOpenedFromAnon': may be the case when:
     * Clicked to the button (open tab with Cognito).
     * Opened UI again ('isCognitoWasOpenedFromAnon' removed).
     * Logined through the Cognito.
     * But the extension will not merge the accounts.
     * Solution: remove 'isCognitoWasOpenedFromAnon' after a few UI openings?
     */
    chrome.storage.sync.remove(['isCognitoWasOpenedFromAnon', 'state']);

    if (!initData) {
        console.log('No token from backend.');
        /** Case: signup was not finished (closed Cognito tab),
         *  adding to the feed from the context menu,
         *  or UI with LOGIN button.
         */
        return;
    }

    console.log('Got init data from the backend:', initData);

    chrome.storage.sync.set({
        'id_token': initData['id_token'],
        'refresh_token': initData['refresh_token'],

        'chars_left': initData['chars_left'],
        'isContinuousPlaying': !!initData['isContinuousPlaying'],
        'playbackRate': initData['playbackRate'] || 1,
        'voices': initData['voices'],

        'updated': Date.now() / 1000 | 0
    });

    const email = _getValueFromIdToken('email', initData);

    /**
    * Edit form at
    * https://docs.google.com/forms/d/1FoHUVP4kNgwMxBucTXzM5Nto2KhoioBUKgRoOGsbOwg/edit
    * Account: intelligentspeakerapp@gmail.com
    *
    * I tested - it enough to set uninstall url only once.
    * Reloading of the extension and restarting of the OS did not reset it.
    */
    chrome.runtime.setUninstallURL(
        uninstallURL + email
    );
    /** TODO also send useragent, locale, storage;
     *  TODO remove on backend from S3 (two buckets), DynamoDB ('Users' and 'history'), Cognito).
     */


    if (location.pathname === '/eventPage.html') {
        // eslint-disable-next-line no-undef
        _initArticles(
            initData['id_token'],
            _whatWeNeedToInit(await _getArticles(), initData['history'])
        );
        return initData;
    }

    chrome.runtime.sendMessage({
        'fromPopup': '_initArticles',
        'id_token': initData['id_token'],
        'history': _whatWeNeedToInit(await _getArticles(), initData['history'])
    });

    return initData;
}

function _whatWeNeedToInit(articlesExisting, articlesAll) {
    return articlesAll.filter(a => !articlesExisting.get(a['utc']));
}


function _getValueFromIdToken(key, storage) {
    console.log('_getValueFromIdToken(): key:', key, 'storage: ', storage);
    const idTokenPayloadStr = storage['id_token'].split('.')[1].replace(/[_-]/g, '/');
    // Without replace() window.atob() can not decode base64
    // We do not store locally JWT after replace() because ApiGateway require original

    let jsonStr = '';
    try {
        jsonStr = atob(idTokenPayloadStr);
    } catch (ex) {
        /* try/catch because received multiple
         * "The string to be decoded is not correctly encoded",
         * but I successfully decoded payload part of id_token from email.
         */
        throw `${ex}, string for decoding: '${idTokenPayloadStr}'`;
    }

    return JSON.parse(jsonStr)[key];
}

// eslint-disable-next-line no-unused-vars
function _insertOrReplace(id, val) {
    return new Promise((resolve, reject) => {
        indexedDB.open('intelligent-speaker').onsuccess = event => {
            const request = event.target.result
                .transaction('articles', 'readwrite')
                .objectStore('articles')
                .put(val, id);
            request.onsuccess = resolve;
            request.onerror = _ => {
                // https://developer.mozilla.org/en-US/docs/Web/API/IDBRequest/onerror
                report('Error put to IndexedDB:' + request.error);
                reject();
            };
        };
    });
}

function _updateIDB(  // eslint-disable-line no-unused-vars
    id,
    key,
    val,
    subkey) {

    return new Promise(resolve => {
        indexedDB.open('intelligent-speaker').onsuccess = event => {
            const request = event.target.result
                .transaction('articles', 'readwrite')
                .objectStore('articles')
                .get(id);

            request.onsuccess = event => {

                if (!event.target.result) {
                    // User log outed and cleared storage
                    return;
                }

                if (typeof subkey !== 'undefined') {  // typeof because subkey may be 0
                    if (!event.target.result[key]) {
                        event.target.result[key] = [];
                    }
                    event.target.result[key][subkey] = val;
                } else {
                    event.target.result[key] = val;
                }

                // Not getFromObjectStore() because source (ObjectStore) will be read only
                event.target.source.put(event.target.result, id);
                resolve();
            };
            request.onerror = event => console.error(event);
        };
    });
}
function _deleteBlob(id, key, subkey) {  // eslint-disable-line no-unused-vars
    return new Promise(resolve => {
        indexedDB.open('intelligent-speaker').onsuccess = event => {
            event.target.result
                .transaction('articles', 'readwrite')
                .objectStore('articles')
                .get(id)
                .onsuccess = event => {
                    delete event.target.result[key][subkey];
                    event.target.source.put(event.target.result, id)
                        .onsuccess = _ => resolve();
                };
        };
    });
}

// eslint-disable-next-line no-unused-vars
function _getArticles(from=0, to=9999999) {
    console.log('_getArticles');

    let i = 0;
    const articles = new Map();
    return new Promise(resolve =>
        indexedDB.open('intelligent-speaker').onsuccess =
        event => event.target.result
            .transaction('articles')
            .objectStore('articles')
            .openCursor(null, 'prev')
            .onsuccess = event => {
                const cursor = event.target.result;
                if (cursor) {
                    if (i++ >= from && articles.size < to) {
                        console.log('_getArticles(), i:', i);
                        articles.set(cursor.key, cursor.value);
                    }
                    cursor.continue();
                } else {
                    resolve(articles);
                }
            }
    );
}

// eslint-disable-next-line no-unused-vars
function getFromObjectStore(id, objectStore='articles') {
    return new Promise(resolve => {
        indexedDB.open('intelligent-speaker')
            .onsuccess = event => {
                const db = event.target.result;
                try {
                    event.target.result
                        .transaction(objectStore)
                        .objectStore(objectStore)
                        .get(id)
                        .onsuccess = event => resolve(event.target.result);
                } catch (ex) {
                    /** Found in automatically emailed client side errors that
                     * "One of the specified object stores was not found"
                     */
                    const objectStoreNames = Array.from(db.objectStoreNames);
                    console.trace();
                    // `Array.from` against `[object DOMStringList]` in the message
                    throw new Error(
                        `getFromObjectStore(): ${ex}, id:${id}, object stores:${objectStoreNames}`
                    );
                }
            };
    });
}
function _deleteArticle(id) {  // eslint-disable-line no-unused-vars
    console.debug('Delete article, id:', id);
    indexedDB.open('intelligent-speaker').onsuccess =
        event => {
            const db = event.target.result;
            try {
                db.transaction('articles', 'readwrite')
                    .objectStore('articles')
                    .delete(id);
            } catch (ex) {
                /** Found in automatically emailed client side errors that
                 * "One of the specified object stores was not found"
                 */
                const objectStoreNames = Array.from(db.objectStoreNames);
                // `Array.from` against `[object DOMStringList]` in the message
                throw `Removing article id:${id}, ex:${ex}, object stores:${objectStoreNames}`;
            }
        };
}


function report(msg) {
    chrome.storage.sync.get(storageSync =>
        _getActiveTab(tabActive => {  // eslint-disable-line no-undef
            fetch(`${stageUrlNew}logextension`, {  // eslint-disable-line no-undef
                'method': 'POST',
                // Without Authorization header - because maybe on some defect will be no token here
                'body': JSON.stringify({
                    'error': msg,
                    'lang': String(window.lang),
                    'url': tabActive.url,
                    'version': chrome.runtime.getManifest().version,
                    'storageCache': window.storageCache,  // TODO also send content from IndexedDB
                    'storageSync': storageSync
                }).replace(/["']/g, '\''),
                // replace for API Gateway against conflicting of " with " of Body Mapping Template
                // replace() ' when this quote conflict with ' of JSON
                'mode': 'no-cors'  // against error in the console during development
            });
        })
    );
    console.error(msg); // Show in JS console.
}


// https://developer.mozilla.org/en-US/docs/Web/API/GlobalEventHandlers/onerror
window.onerror = (message, source, lineno, colno, error) => {  // eslint-disable-line no-unused-vars
    let msg = 'message:`' + message + '`';
    if (error && error.stack) {  // No error arg when throw text
        msg += ', stacktrace:' + error.stack;
        /** For Chromium is enough error.stack - error message will be here,
         *  but Firefox does not error message in error.stack
         */
    }
    report(msg);
};
/**
 * https://developer.mozilla.org/en-US/docs/Web/API/WindowEventHandlers/onunhandledrejection
 */
window.onunhandledrejection = e => {
    const prefix = 'Message of rejected Promise:';
    if (e.reason.message) {  // Native JS error object
        let msg = prefix + e.reason.message;
        if (e.reason.stack) {
            msg += '`, stacktrace:' + e.reason.stack;
        }
        report(msg);
    } else {
        report(prefix + '(reported from e.reason)' + e.reason);
    }
};
