/* global stageUrlNew,
 _updateIDB,
 getFromObjectStore,
 _getRandomString,
 _getArticles,
 report,
 CHARS_IN_MINUTE,
 _gaq
 */

// cache - for 1) speed 2) against limit of read 3) against one more callback indentation */
let storageCache,
    messages,
    lang,
    nodes = {};
// TODO rename to gStorageCache

// Long, start ASAP
_fillListAudio();

chrome.storage.sync.get(resp => {
    storageCache = resp;


    /** Navigator.language return 'en-US' or 'en', 'fr' or 'fr_FR'
     * I tested this: on my main profile of the Chrome was returned 'en',
     * on 'Another' profile was returned 'en-US',
     * also one user wrote email that in JS console see error about fr_FR
     *
     *  TODO store language after install?
     *  for backward compatibility - set language to the storage after update.
     *  It will make code simpler but less robust?
     */
    lang = storageCache['lang'] || _getLang();  // eslint-disable-line no-undef
    _getMessages(lang)  // eslint-disable-line no-undef
        .then(resp => {
            messages = resp['messages'];
            dispatchEvent(new Event('messagesIsReady'));
            if (resp['isFallback'])
                lang = 'en';
        })
        .then(starter);
});

function starter() {

    _localize(storageCache);

    const isUserLogouted = !('chars_left' in storageCache);
    if (isUserLogouted) {
        UILogout();

        if (storageCache['state']) {
            _progressShow();
            // eslint-disable-next-line no-undef
            _init(storageCache['state']).then(initData => {
                if (initData) {
                    location.reload();
                } else {
                    _progressHide();
                }
            });
        }
        return;
    }

    if (storageCache['isCognitoWasOpenedFromAnon']) {
        _progressShow();
        // eslint-disable-next-line no-undef
        _init(_getValueFromIdToken('email', storageCache)).then(initData => {
            if (initData) {
                location.reload();
            } else {
                _progressHide();
            }
        });
    }

    initOpening();
    initMessageListener();
    chrome.runtime.sendMessage({'fromPopup': 'getStatus'});
    _podcastsButtonHideOrActivate();
    _accountButtonHideOrActivate();
    initEventListeners();
    /* TODO once in a while check cache, even play without sound every chunk -
     * for checking Promise after play() -
     * maybe in IndexedDB only link without blob - I faced with such case.
     */

    _initGoogleAnalytics();
}

function _initGoogleAnalytics() {
    _gaq.push(['_setAccount', 'UA-105882536-1']);
}

/** February 2018: Google Chrome i18n works only for manifest and CSS, see
 * https://bugs.chromium.org/p/chromium/issues/detail?id=115800
 * TODO for faster replacing - iterate through the hardcoded list of nodes?
 *
 * TODO instead of DOM manipulating - sh will generate html?
 * Pros: faster loading without visible changing of the texts in html.
 * Cons: altering of the language will nagigator.reload() and blink the UI.
 * Little longer install, dozen of html files.
 */
function _localize() {
    const _localizeNode = (node, placeholder) =>
        node.innerText = _getMessage(
            (placeholder || node.innerText).replace(/__MSG_(\w+)__/, '$1')
        );

    // Against consuming much blocking time - localize visible strings first
    nodes['search'] = document.getElementById('search');
    nodes['search'].placeholder = _getMessage('placeholder_search');
    _localizeNode(document.getElementById('text-playback-rate'));
    _localizeNode(document.getElementById('empty'));

    nodes['login-or-logout'] = document.getElementById('login-or-logout');
    _localizeNode(
        nodes['login-or-logout'].children[1],
        // eslint-disable-next-line no-undef
        isAnon() ? '__MSG_login__' : '__MSG_logout__'
    );

    setTimeout(_ => {
        for (const node of document.querySelectorAll('*')) {
            // TODO what if page title contains '__MSG__someShit__'?
            if (!node.childElementCount && node.innerText.match(/^__MSG_\w+__$/))
                _localizeNode(node);
        }
    }, 99);
}


/**
 * "Overriden" chrome.i18n.getMessage()
 * https://developer.chrome.com/extensions/i18n#method-getMessage
 * because we need to have the ability to set lang implicitly.
 */
function _getMessage(messageName, substitutions) {
    if (substitutions)
        return messages[messageName]['message'].replace(
            /\$(\w+)\$/,
            (_, group) => substitutions[
                messages[messageName]['placeholders'][group]['content'].replace('$', '') - 1
            ]
        );
    else
        return messages[messageName]['message'];
}

/** Build nodes object because it is faster to select node only once.
 *  Link on the start because it is simpler :) and cheaper -
 *  at moment of selectors we do not have feed list (simpler DOM).
 */
function initOpening() {
    console.time('initOpening');

    nodes = {...{
        'logo' : document.getElementById('logo'),
        'hamburger' : document.getElementById('hamburger'),

        'drawer' : document.querySelector('.mdc-drawer'),

        'upgrade': document.getElementById('upgrade'),

        'listFeed' : document.getElementById('list-feed'),

        'addToFeed' : document.getElementById('add-to-feed'),
        'openFile' : document.getElementById('open-file'),

        'player' : document.getElementById('player'),
        'playPause' : document.getElementById('play-pause'),
        'prevChunk' : document.getElementById('prev-chunk'),
        'nextChunk' : document.getElementById('next-chunk'),

        'continuousPlaying' : document.getElementById('continuous-playing'),

        'playbackRateWrapper' : document.getElementById('playback-rate-wrapper'),
        'playbackRates' : document.getElementsByName('playbackRate'),
        'playbackRateLabels' : document.getElementsByTagName('label'),

        'snackbar' : document.querySelector('intelligentspeaker--snackbar'),
        'dialog' : document.getElementById('mdc-dialog'),
    }, ...nodes};

    if (storageCache['id_token']) {
        // id_token is missing on the first opening (until receiving)

        _keepTokenFresh(storageCache).then(  // eslint-disable-line no-undef
            storageWithMaybeRefreshedToken => {
                storageCache = storageWithMaybeRefreshedToken;

                // eslint-disable-next-line no-undef
                if (_getValueFromIdToken('email', storageCache).endsWith('@anonymous')) {
                    _updateUser(storageCache['id_token']);
                } else {
                    _syncHistory(storageCache['updated'], storageCache['id_token']);
                    _updateUser(storageCache['id_token'], true);
                }
            });
    }

    _getActiveTab(tabActive => {  // eslint-disable-line no-undef
        if (
            tabActive.url.match(
                /^https:\/\/(chrome.google.com\/webstore|addons.mozilla.org|addons.opera.com)/
                // TODO handle Edge
            ) || tabActive.url.match(/^chrome|^about/)
        )
            _disableUI();
    });
    if (storageCache['isContinuousPlaying'] === true)
        nodes['continuousPlaying'].classList.add('true');

    nodes['playbackRateWrapper'].querySelector(
        `[value="${storageCache['playbackRate']}"]`
    ).checked = true;

    // Execute it later because this screen is not visible by default
    setTimeout(_ => {
        _setPersonalization();
    }, 99);

    if (storageCache['isSubscribed']) {
        nodes['upgrade'].classList.add('hidden');
    } else {
        nodes['upgrade'].onclick = _ => _openSitePage('pricing');
    }

    console.timeEnd('initOpening');
}

function _setPersonalization() {
    nodes['avatar'] = document.getElementById('avatar'),
    nodes['avatarPlaceholder'] = document.getElementById('avatar-placeholder'),
    nodes['username'] = document.getElementById('username'),
    nodes['email'] = document.getElementById('email'),
    nodes['availableTime'] = document.getElementById('available-time');

    if (storageCache['id_token']) {
        // eslint-disable-next-line no-undef
        const picture = _getValueFromIdToken('picture', storageCache);
        if (picture) {
            if (picture[0] === '{') {
                const pictureJSON = JSON.parse(picture);
                if (pictureJSON['data']) {
                    nodes['avatar'].src = pictureJSON['data']['url'];
                } else {
                    nodes['avatar'].src = pictureJSON['source'];
                }
            } else {
                nodes['avatar'].src = picture;
            }
            nodes['avatarPlaceholder'].classList.add('hidden');
        }

        // eslint-disable-next-line no-undef
        const givenName = _getValueFromIdToken('given_name', storageCache);

        // eslint-disable-next-line no-undef
        const familyName = _getValueFromIdToken('family_name', storageCache);

        nodes['username'].innerText = [givenName, familyName].join(' ');

        // eslint-disable-next-line no-undef
        const email = _getValueFromIdToken('email', storageCache);

        // TODO i18n
        nodes['email'].innerText = isAnon() ? 'Anonymous account' : email;

    } else {
        nodes['username'].innerText = 'Anonymous'; // TODO localize?
    }

    nodes['availableTime'].innerText =
        _minutesToHumanText(storageCache['chars_left'] / CHARS_IN_MINUTE);
}

function _podcastsButtonHideOrActivate() {
    nodes['emailPodcastLink'] = document.getElementById('email-podcast-link');
    if (isAnon()) {
        nodes['emailPodcastLink'].remove();
        return;
    }

    nodes['emailPodcastLink'].onclick = function() {
        nodes['drawer'].classList.remove('mdc-drawer--open');
        nodes['settingsView'].classList.remove('show');
        this.disabled = true;
        _progressShow();

        _keepTokenFresh(storageCache)  // eslint-disable-line no-undef
            .then(tokensFresh => emailPodcastLink(tokensFresh));
    };
}
function _accountButtonHideOrActivate() {
    nodes['accountPageLink'] = document.getElementById('account-page-link');
    if (isAnon()) {
        nodes['accountPageLink'].remove();
        return;
    }
    nodes['accountPageLink'].onclick = _ => _openSitePage('account');
}

function _getBrowserName() {
    let name = navigator.userAgent.match(/Chrome|OPR|Vivaldi|YaBrowser|Firefox/g)
        .pop();
    if (name === 'OPR') {
        name = 'Opera';
    }
    return name;
}

function _fillInvisiableNodesAndEvents() {
    nodes['feedback'] = document.getElementById('feedback'),
    nodes['ideas'] = document.getElementById('ideas'),
    nodes['chat'] = document.getElementById('chat');
    nodes['guide'] = document.getElementById('guide');

    nodes['settingsButton'] = document.getElementById('settings-button'),
    nodes['settingsView'] = document.getElementById('settings-view'),

    nodes['backdrop'] = document.getElementById('mdc-dialog__backdrop'),

    nodes['languages'] = document.getElementById('languages');

    // eslint-disable-next-line no-undef
    const email = storageCache['id_token'] ? _getValueFromIdToken('email', storageCache)
        :'Anonymous';

    nodes['feedback'].onclick = _ => {
        const url = 'https://docs.google.com/forms/d/e/' +
            '1FAIpQLSdPh9fuy1P3zdbXO6knjCRrinzoKzTdRGu2wqdxvVIpZTblMw/viewform' +
            '?usp=pp_url' +
            '&entry.337702430' +
            '&entry.2137746123=' + email +
            '&entry.1853332692=' + _getBrowserName(); // eslint-disable-line no-undef
        chrome.tabs.create({'url': url});
    };

    nodes['ideas'].onclick = _ => {
        const url =
            // eslint-disable-next-line max-len
            'https://docs.google.com/forms/d/e/1FAIpQLSejGIENm_CEigc0cmIGW4-6xPVhgxLYxWqGRCNDortizEmS_g/' +
            'viewform?usp=pp_url' +
            '&entry.826480340' +
            '&entry.968924808=' + email;
        chrome.tabs.create({'url': url});
    };


    nodes['chat'].onclick = _ => {
        // Not short url 'm.me/IntelligentSpeaker' against HTTP 301 redirect
        chrome.tabs.create({'url': 'https://www.facebook.com/messages/t/IntelligentSpeaker'});
    };

    nodes['guide'].onclick = _ => {
        chrome.tabs.create({'url':
            'https://www.youtube.com/watch?v=GkN1WQvXREU&list=PLiccyGSihaDdmcdYxLiUkxuCOe-9SnBRX'});
    };

    nodes['settingsButton'].onclick = _ => {
        nodes['settingsView'].classList.toggle('show');
        nodes['settingsButton'].classList.toggle('show');

        // Add event listener only when menu opened - for faster general UI opening
        document.querySelector('#settings-view #close').onclick = _ => {
            nodes['settingsView'].classList.remove('show');
            nodes['settingsButton'].classList.remove('show');
        };

        _gaq.push(['_trackEvent', 'ui', 'settings']);
    };

    nodes['backdrop'].onclick = function() {
        _closeDialog();
    };

    nodes['login-or-logout'].onclick = _ => {
        if (isAnon()) {
            chrome.storage.sync.set({'isCognitoWasOpenedFromAnon': true});

            // eslint-disable-next-line no-undef
            openCognitoPage(_getValueFromIdToken('email', storageCache));
        } else {
            logout();
        }
    };

    (function initSettings() {

        const listSetDefaultPosition = scrollable => {
            console.debug(
                'Now will move scrollable part of the dropdown to the init position, ',
                'scrollable:', scrollable
            );

            if (Array.from(scrollable.children)
                .findIndex(node => node.classList.contains('checked')) < 10) {

                scrollable.style.top = '20px';  // Hardcode but fast and simple
            } else {
                scrollable.style.top = '-280px';
            }

        };

        (function initLanguagesDropdown() {
            const scrollable = nodes['languages'].querySelector('section');
            const nodeLang = scrollable.querySelector(`[data-value=${lang}`);
            nodeLang.classList.add('checked');
            nodes['languages'].dataset['text'] = nodeLang.innerText;

            listSetDefaultPosition(scrollable);

            nodes['languages'].onclick = function(event) {
                if (event.target.id !== 'languages' && !this.classList.contains('open')) {
                    return;
                    // Clicked on empty space below the dropdowns
                }

                const isOpening = this.classList.toggle('open');
                document.querySelectorAll('.multiselect')
                    .forEach(node => node.classList.remove('open'));

                if (event.target.classList.contains('checked')) {
                    return;
                }

                addEventListener(
                    'wheel',
                    scrollEvent => {
                        if (!this.classList.contains('open')) return;
                        _onwheel(scrollEvent, scrollable);
                    },
                    {'passive': true}  // Chrome warning without it - for performance
                );

                // Do not bubble click event to the window handler that will close the dropdown
                event.stopPropagation();
                window.onclick = _ => {
                    if (this.classList.contains('open')) {
                        _closeDropdown();
                        listSetDefaultPosition(scrollable);
                        nodes['settingsView'].classList.remove('doNotScroll');
                    }
                };

                if (isOpening) {
                    nodes['settingsView'].classList.add('doNotScroll');
                    return;
                }

                const lang = event.target.dataset['value'];
                chrome.storage.sync.get('lang', resp => {
                    if (resp['lang'] !== lang) {

                        chrome.storage.sync.set({'lang': lang});

                        chrome.contextMenus.removeAll();
                        chrome.runtime.sendMessage({'fromPopup': 'buildContextMenu'});

                        _gaq.push(['_trackEvent', 'ui', 'languageui-change']);
                        location.reload();
                    }
                });

            };
        })();

        (function initVoicesUI() {

            document.getElementById('label-voice').onclick = event => {
                event.target.classList.toggle('show');
            };
            _multiselectRestoreView();

            document.getElementById('settings-view').onclick = event => {
                console.log(event);

                if (!_isThisNodeWithVoice(event.target) &&
                    !event.target.classList.contains('dropdown') ||

                    event.target.classList.contains('dropdown') &&
                   !event.target.classList.contains('open')
                ) {
                    _closeDropdown();
                }

                if (event.target.classList.contains('multiselect')) {

                    const multiselect = event.target.children[0];
                    console.log('multiselect.innerText:', multiselect.innerText);
                    multiselect.style.top = event.y - 15 + 'px';
                    console.log('multiselect.style.top before:', multiselect.style.top);

                    _updateDropdownPositionAfterReflow(multiselect, event.y);

                    event.target.classList.add('open');
                    event.target.previousElementSibling.classList.add('highlight');

                    const wheelEventListener = scrollEvent => {
                        console.log('onwhell');
                        if (!event.target.classList.contains('open')) {
                            return;
                        }
                        const countOfVoicesInThisMultiselect =
                            multiselect.querySelectorAll('div').length;
                        if (countOfVoicesInThisMultiselect > 5) {
                            _onwheel(scrollEvent, multiselect);
                        }
                    };
                    addEventListener(
                        'wheel',
                        wheelEventListener,
                        {'passive': true}  // Chrome warning without it - for performance
                    );

                    nodes['settingsView'].classList.add('doNotScroll');

                    _setEventsOnVoicesMultiselect(event.target);

                }
            };

        })();

        (function initBreatingToggle() {
            const breathing = document.getElementById('breathing');
            if (storageCache['breathing'] === false) {
                breathing.classList.remove('on');
            }
            breathing.onclick = function() {
                const val = this.classList.toggle('on');
                chrome.storage.sync.set({'breathing': val});
                storageCache['breathing'] = val;

                _gaq.push(['_trackEvent', 'ui', 'breath']);
            };
        })();

    })();
}

function isAnon() {
    if (storageCache['id_token']) {
        // eslint-disable-next-line no-undef
        return _getValueFromIdToken('email', storageCache).endsWith('@anonymous');
    } else {
        /** id_token is not exists only when user installed the extension and opened it fast -
         *  when anonymous id_token still in the process of receiving.
         */
        return true;
    }
}


function logout() {
    chrome.runtime.sendMessage({'fromPopup': 'stopPlaying'});
    chrome.storage.sync.clear();

    // http://docs.aws.amazon.com/cognito/latest/developerguide/logout-endpoint.html
    fetch('https://auth.intelligent-speaker.com/logout' +
                    '?client_id=' + clientId +  // eslint-disable-line no-undef
                    '&logout_uri=https://api.intelligent-speaker.com/logout', {
        'mode': 'no-cors'
    });

    indexedDB.open('intelligent-speaker').onsuccess =
        event => event.target.result
            .transaction('articles', 'readwrite')
            .objectStore('articles')
            .clear();
    // TODO if still caching - after login user will see the history of previous user

    UILogout();

    chrome.runtime.sendMessage({'fromPopup': 'stopPlaying'});

    _gaq.push(['_trackEvent', 'ui', 'logout']);
}
function UILogout() {
    const nodeLogin = document.getElementById('login');
    for (const node of document.body.getElementsByTagName('*')) {
        node.style.display = 'none';
    }

    nodeLogin.innerText = _getMessage('login');
    nodeLogin.style.display = 'block';

    nodeLogin.onclick = _ => {
        const state = _getRandomString();
        chrome.storage.sync.set({'state': state});

        // eslint-disable-next-line no-undef
        openCognitoPage(state);
    };
}

function _closeDropdown() {
    const dropdown = document.querySelector('.dropdown.open');
    if (dropdown) {
        dropdown.classList.remove('open');
        dropdown.previousElementSibling.classList.remove('highlight');
    }
    nodes['settingsView'].classList.remove('doNotScroll');
}

function _updateDropdownPositionAfterReflow(multiselect, eventY) {
    setTimeout(_ => {
        // setTimeout is necessary, see https://stackoverflow.com/questions/26813480
        const bottom = multiselect.getBoundingClientRect().bottom;
        if (bottom + 20 > innerHeight) {
            multiselect.style.top = 15 + (multiselect.clientHeight > eventY ? 0 :
                eventY - multiselect.clientHeight) +
                'px';
        }
    }, 60);
}

function _isThisNodeWithVoice(node) {
    return node.dataset['voice'];
}

function _onwheel(scrollEvent, scrollable) {
    const offset = 20;
    const bottomLimiter = innerHeight - scrollable.offsetHeight - offset;
    const val = +scrollable.style.top.replace('px', '');
    const topLimiter = 50;
    const scrollToBottom = scrollEvent.deltaY > 0;
    const scrollToTop = scrollEvent.deltaY < 0;
    if (val > topLimiter && scrollToTop ||
        val < bottomLimiter && scrollToBottom) return;

    // https://developer.mozilla.org/en-US/docs/Web/API/WheelEvent/deltaMode
    // https://github.com/w3c/uievents/issues/181#issuecomment-392648065
    let valNew = val;
    if (scrollEvent.deltaMode === 0) {
        valNew -= scrollEvent.deltaY;
    } else if (scrollEvent.deltaMode === 1) {
        valNew -= scrollEvent.deltaY * 10;
    } else if (scrollEvent.deltaMode === 2) {
        valNew -= scrollEvent.deltaY * 100;
    }

    if (valNew < bottomLimiter) {
        console.log('bottomLimiter:', bottomLimiter);
        valNew = bottomLimiter;
    }
    if (valNew > topLimiter) {
        console.log('topLimiter:', topLimiter);
        valNew = topLimiter;
    }
    console.log('valNew:', valNew);
    requestAnimationFrame(_ => {
        scrollable.style.top = valNew + 'px';
    });
}

function _multiselectRestoreView() {
    chrome.storage.sync.get(storageSync => {
        for (const multiselectNode of document.getElementsByClassName('multiselect')) {
            const voiceIDsForLang = (storageSync['voices'] || {})[multiselectNode.dataset['lang']];
            if (voiceIDsForLang) {
                multiselectNode.dataset['text'] =
                    _getVoiceStringsForView(multiselectNode, voiceIDsForLang);
                _multiselectRestoreCheckeds(multiselectNode, voiceIDsForLang);
                continue;
            }
            const nodeVoiceFirstInList = multiselectNode.querySelector('div');
            multiselectNode.dataset['text'] = nodeVoiceFirstInList.innerText;
        }
    });
}
function _getVoiceStringsForView(multiselectNode, voiceIDsForLang) {
    return voiceIDsForLang
        .map(voiceID => multiselectNode.querySelector(`[data-voice="${voiceID}"]`)
            .innerText.match(/: (.+)/)[1])
        .join(', ');
}
function _multiselectRestoreCheckeds(multiselectNode, voices) {
    voices.forEach(voice => {
        multiselectNode.querySelector(`[data-voice="${voice}"]`).classList.add('checked');
    });
}
function _setEventsOnVoicesMultiselect(multiselect) {
    for (const node of multiselect.querySelectorAll('div')) {
        const voiceID = node.dataset['voice'];
        node.onclick = event => {
            console.log(event);
            if (event.layerX > 240) {
                voicePreview(node, voiceID);
                _gaq.push(['_trackEvent', 'ui', 'voice-preview']);
                return;
            }

            const isAdded = node.classList.toggle('checked');
            const lang = event.target.parentElement.parentElement.dataset['lang'];
            _saveVoiceConfAndUpdateView(lang, voiceID, isAdded, multiselect);
            _gaq.push(['_trackEvent', 'ui', 'voice-change']);
        };
    }
}
function voicePreview(node, voiceID) {
    let isArticlePlayingWasPaused = false;
    if (nodes['playPause'].classList.contains('playing')) {
        isArticlePlayingWasPaused = true;
        chrome.runtime.sendMessage({'fromPopup': 'pause'});
    }
    dispatchEvent(new Event('voicePreviewStarted'));

    const _play = (url, id) => {
        const audio = new Audio(url);
        audio.play();
        node.classList.add('playing');
        addEventListener('voicePreviewStarted', _ => {
            audio.pause();
            node.classList.remove('playing');
        });
        audio.onended = _ => {
            node.classList.remove('playing');

            if (isArticlePlayingWasPaused) {
                setTimeout(
                    _ => chrome.runtime.sendMessage({'fromPopup': 'resume'}),
                    300
                );
            }

            if (id) {
                chrome.runtime.sendMessage({'fromPopup': '_cacheVoicePreview', id});
            }
        };
    };

    const lang = node.parentElement.parentElement.dataset['lang'];
    const id = (node.dataset['locale'] || lang) + '-' + voiceID;
    getFromObjectStore(id, 'voice-preview').then(resp => {
        if (resp) {
            _play(URL.createObjectURL(resp['blob']));
        } else {
            // PC was turned off during fetch after install
            _play(`https://intelligent-speaker.com/voice-preview/${id}.ogg`, id);
        }
    });
}

/** For migrating users from old version of extension,
/*  TODO remove this function in the year 2020.
 */
function __convertFromDeprecatedStringIfNeeded(val) {
    return typeof val === 'string' ? [val] : val;
}
function _saveVoiceConfAndUpdateView(lang, voiceID, isAdded, multiselectNode) {
    chrome.storage.sync.get(storage => {

        const voices = storage['voices'] || {};

        const voiceIDsForLang = __convertFromDeprecatedStringIfNeeded(voices[lang]) || [];

        if (isAdded) {
            voiceIDsForLang.push(voiceID);
        } else {
            const index = voiceIDsForLang.findIndex(val => val === voiceID);
            voiceIDsForLang.splice(index, 1);
        }

        multiselectNode.dataset['text'] =
            _getVoiceStringsForView(multiselectNode, voiceIDsForLang) ||
            multiselectNode.querySelector('div').innerText;

        if (voiceIDsForLang.length) {
            voices[lang] = voiceIDsForLang;
        } else {
            delete voices[lang];
        }

        chrome.storage.sync.set({voices});
        _setSetting('voices', voices);
    });
}

function _disableUI() {
    nodes['addToFeed'].classList.add('hidden');
}


function initMessageListener() {
    chrome.runtime.onMessage.addListener(fromEventPage => {
        console.log('Event from content script or event page:', fromEventPage);
        if (fromEventPage['toPopup']) {

            // Case: reopening of the popup while fetching of the chunks still in progress
            if (fromEventPage['toPopup']['isPreparingPlaying']) {
                nodes['playPause'].disabled = true;
                nodes['listFeed'].classList.add('disabled');
            }

            // Do not block button `Add to feed` - I tested on very slow connection
            // (custom throttling preset in Chrome)
            // and it is ok to add a pages to the feed
            // while previous still adding -
            // they will be resolved correctly
            if (fromEventPage['toPopup']['stillAddingToFeed']) {
                fromEventPage['toPopup']['stillAddingToFeed'].forEach(msg => {
                    _progressShow();

                    /* 0 and 1 because Map was converted into array,
                     * map used because we need to remove item by id.
                     */
                    const id  = msg[0];
                    const obj = msg[1];

                    if (document.getElementById(id)) {
                        document.getElementById(id).classList.add('addingToFeed');
                        return;
                    }

                    _persistItemFeed(
                        id,
                        obj['favIconUrl'],
                        obj['url'],
                        obj['title'],
                        true,
                        true
                    );

                });
            }

            if (fromEventPage['toPopup']['playingStarted']) {
                nodes['listFeed'].classList.remove('disabled');

                nodes['playPause'].disabled = false;
                nodes['playPause'].classList.add('playing');

                const node = document.getElementById(
                    fromEventPage['toPopup']['audioCurrentFeedEntityId']
                );
                node.classList.remove('new');
            }

            if (fromEventPage['toPopup']['audioCurrentFeedEntityId']) {
                _executeNowOrWhenNodesAppears(
                    _ => [document.querySelector('.current'),
                        document.getElementById(
                            fromEventPage['toPopup']['audioCurrentFeedEntityId']
                        )],
                    (itemFeedCurrent, itemFeedNew) => {
                        itemFeedNew.classList.add('current');

                        if (itemFeedCurrent && itemFeedCurrent.id !== itemFeedNew.id)
                            itemFeedCurrent.classList.remove('current', 'playing');

                        if (fromEventPage['toPopup']['isPaused'] === false) {
                            itemFeedNew.classList.add('playing');
                            nodes['playPause'].classList.add('playing');

                        } else if (fromEventPage['toPopup']['playingStarted']) {
                            itemFeedNew.classList.add('playing');
                        }
                    }
                );
            }

            if (fromEventPage['toPopup']['index'] > 0) {
                nodes['prevChunk'].disabled = false;
                nodes['playPause'].disabled = false;
            } else if (fromEventPage['toPopup']['index'] === '0') {
                nodes['prevChunk'].disabled = true;
            }

            if (fromEventPage['toPopup']['hasNextChunk'] === true) {
                nodes['playPause'].disabled = false;
                nodes['nextChunk'].disabled = false;
            } else if (fromEventPage['toPopup']['hasNextChunk'] === false) {
                nodes['nextChunk'].disabled = true;
            }

            if (fromEventPage['toPopup']['isPaused']) {
                _executeNowOrWhenNodesAppears(
                    _ => [document.querySelector('.current')],
                    node => node.classList.remove('playing')
                );
                nodes['playPause'].classList.remove('playing');
            }
            if (fromEventPage['toPopup']['isPaused'] === false) {
                nodes['playPause'].classList.add('playing');
            }

            if (fromEventPage['toPopup']['playingFinished']) {

                const itemFeed = document.querySelector('.current');
                // Feed item not highlighted when playing current page
                if (itemFeed) {
                    if (fromEventPage['toPopup']['removeStubFromFeed']) {
                        itemFeed.remove();
                    } else {
                        itemFeed.classList.remove('current', 'playing');
                        console.log('feed item: after unhighlighting:', itemFeed);
                    }
                }

                nodes['listFeed'].classList.remove('disabled');
                // When playing was not started

                nodes['prevChunk'].disabled = true;

                /** Needed when playing finished without un-disabled:
                 *  when some error on server or offline.
                 *  Disabled on starting.
                 */
                nodes['playPause'].disabled = false;

                nodes['playPause'].classList.remove('playing');

                nodes['nextChunk'].disabled = true;
            }

            if (fromEventPage['toPopup']['errorAddingToFeed']) {
                _progressHide();
                _executeNowOrWhenNodesAppears(
                    _ => [document.querySelector('.addingToFeed')],
                    node => node.remove()
                );

                if (fromEventPage['toPopup']['reason']) {
                    _showSnackbar(fromEventPage['toPopup']['reason']);
                }
            }

            if (fromEventPage['toPopup'] === 'addedToFeed') {
                if (fromEventPage['hashArticle']) {
                    _showSnackbar(_getMessage('snackbar_addedToFeed'));
                } else {
                    // Adddd in offline TODO
                    //node.classList.add('addedOffline');
                    //TODO css; TODO block playing before sync
                    //node.dataset['tooltip'] = _getMessage('tooltip_thisItemOffline');
                    //
                    // TODO add this key into i18n jsons
                    //_showSnackbar(_getMessage('snackbar_addedToFeedOffline'));
                }

                _progressHide();

                _executeNowOrWhenNodesAppears(
                    _ => [document.querySelector('.addingToFeed')],
                    node => node.classList.remove('addingToFeed')
                );

                /** When user added first page to the feed.
                 *  Checking added against the case when page preparing for playing -
                 *  play/pause button must remained disabled.
                 */
                if (document.getElementsByClassName('item-feed').length === 1) {
                    nodes['playPause'].disabled = false;
                }
            }
            if (fromEventPage['toPopup']['cachedChunkId']) {
                _executeNowOrWhenNodesAppears(
                    _ => {
                        const item =
                            document.getElementById(fromEventPage['toPopup']['cachedChunkId']);
                        return item ? [item.querySelector('.tooltip')] : [];
                    },
                    node => {
                        if (isNaN(node.innerText[node.innerText.length - 1])) {
                            const lenChunks = fromEventPage['toPopup']['lenChunks'];

                            if (lenChunks > 1) {
                                node.innerText =
                                    _getMessage('tooltip_download_caching') + ` 1 / ${lenChunks}`;
                            } else {
                                node.parentElement.classList.remove('processing');
                                node.innerText = _getMessage('tooltip_download');
                            }
                        } else {
                            const count = fromEventPage['toPopup']['count'];
                            node.innerText = node.innerText.replace(/(\d+).+?(\d+)/,
                                (_, now, max) => {
                                    if (now < max - 1) {
                                        return `${count} / ${max}`;
                                    }
                                    node.parentElement.classList.remove('processing');
                                    const text = _getMessage('tooltip_download');
                                    setTimeout(_ => node.innerText = text, 0);
                                }
                            );
                        }
                    }
                );
            }

            if (fromEventPage['toPopup']['removeStub']) {
                nodes['listFeed'].firstElementChild.remove();
            }

            if (fromEventPage['toPopup']['_fillListAudio']) {
                _fillListAudio();
            }

        }
    });
}
function _executeNowOrWhenNodesAppears(getElements, callback) {
    const elements = getElements();
    if (elements.some(resp => resp)) {
        callback(...elements);
    } else {
        new MutationObserver(function() {
            const elements = getElements();
            if (elements.some(resp => resp)) {
                this.disconnect();
                callback(...elements);
            }
        }).observe(nodes['listFeed'], {'childList': true});
    }
}

function initEventListeners() {

    console.time('initEventListeners');

    nodes['search'].onkeyup = function() {
        const nodesToHide = [];
        const nodesToShow = [];
        for (const node of nodes['listFeed'].children)
            if (node.innerText.toLowerCase().includes(this.value.toLowerCase()))
                nodesToShow.push(node);
            else
                nodesToHide.push(node);

        /* Why this arrays instead of add/remove inside if/else?
         * Because with naive approach in JS console of Chromium you will see
         * `Forced reflow while executing JavaScript took <some number if ms more than 50ms>`
         * In naive approach time eated by rerendering for every node
         * and I see the difference.
         * https://stackoverflow.com/questions/41218507/violation-long-running-javascript-task
         */
        nodesToHide.forEach(node => node.classList.add('hidden'));
        nodesToShow.forEach(node => node.classList.remove('hidden'));

        _gaq.push(['_trackEvent', 'ui', 'search']);
    };

    nodes['logo'].onclick = _ => _openSitePage();

    nodes['hamburger'].onclick = function() {
        _gaq.push(['_trackEvent', 'ui', 'drawer']);
        nodes['drawer'].classList.add('mdc-drawer--open');

        setTimeout(_fillInvisiableNodesAndEvents, 300);

        /** Init on drawer opening but not the Settings because on open of Settings
         *  user should not see CSS animation of toggles.
         */
    };
    // Close drawer on click to the grayed zone
    nodes['drawer'].onclick = function(mouseEvent) {
        if (mouseEvent.target.classList.contains('mdc-drawer--temporary'))
            this.classList.remove('mdc-drawer--open');
    };

    /** TODO case: user trying to add the same page -
     *  but note that text can differ on the same url and the same title.
     */
    nodes['addToFeed'].onclick = function() {
        console.log('add to feed - event listener on button');

        // For the first added item, but I off the IF
        nodes['listFeed'].classList.remove('empty');

        this.disabled = true;  // Against second click

        _getActiveTab(tabActive => {  // eslint-disable-line no-undef
            // Date.now() returns milliseconds - stripped to seconds and rounded
            console.log('tabActive.favIconUrl:', tabActive.favIconUrl);
            chrome.tabs.detectLanguage(lang => {
                if (lang === 'und') {
                    // I faced on Gmail when selected phrase in English - check lang of selected
                    // TODO try to use English anyway?
                    // TODO log this page for review - send url to the special lambda?
                    _showSnackbar(_getMessage('snackbar_error_languageOfThePageIsUndefined'));
                    return;
                }
                // eslint-disable-next-line no-undef
                if (!chrome.runtime.lastError && !_isLangSupported(lang)) {
                    /** chrome.runtime.lastError: "Cannot determine language" -
                     *  faced with this in Opera for normal pages, looks like their defect.
                     */
                    _showSnackbar(_getMessage('languageNotSupported'));
                    return;
                }

                const time = _getNextUTCId();

                /** It is more reliable to take ico from chrome API instead of content script
                /*  because ico can be set in code in a few ways, from js too.
                 */
                chrome.runtime.sendMessage({
                    'fromPopup': 'addToFeed',
                    'time': time,
                    'url': tabActive.url,
                    'favIconUrl': tabActive.favIconUrl,
                    'title': tabActive.title,
                    'lang': lang
                });

                _progressShow();

                _persistItemFeed(
                    time,
                    tabActive.favIconUrl,
                    tabActive.url,
                    tabActive.title,
                    true,
                    true
                );
            });

        });
        _gaq.push(['_trackEvent', 'ui', 'addtofeed']);
    };
    nodes['openFile'].onclick = function() {
        // For the first added item, but I off the IF
        nodes['listFeed'].classList.remove('empty');

        const input = document.querySelector('[type=file]');
        input.click();
        input.onchange = function() {
            _progressShow();

            const file = this.files[0];
            console.debug(file);

            const time = _getNextUTCId();

            const icos = {
                'text/plain':      'https://intelligent-speaker.com/images/txt.svg',
                'application/pdf': 'https://intelligent-speaker.com/images/pdf.svg',
                'application/vnd.openxmlformats-officedocument.wordprocessingml.document':
                    'https://intelligent-speaker.com/images/docx.svg'
            };

            _persistItemFeed(
                time,
                icos[file.type],
                '',
                file.name,
                true,
                true
            );

            if (file.type === 'text/plain') {
                /* Max filesize for Lambda sync is 6MB,
                 * such amount of plaintext is more than 24 hours -
                 * out limit in premium plan,
                 * so will be faster to not use intermediate S3 in such case.
                 * see https://docs.aws.amazon.com/lambda/latest/dg/limits.html
                 **/

                const reader = new FileReader();
                reader.readAsText(file, 'UTF-8');
                reader.onload = event => {
                    const stringForVoicing = event.target.result;
                    chrome.i18n.detectLanguage(stringForVoicing, resp => {
                        console.debug('Language detection of file:', resp);
                        chrome.runtime.sendMessage({
                            time,
                            'title': file.name,
                            'favIconUrl': icos[file.type],
                            stringForVoicing,
                            'lang': resp['languages'][0]['language']
                        });
                    });
                };

            } else {
                const prefix = _getRandomString() + '/' + file.name;
                fetch('https://input-files-for-tts.s3.amazonaws.com/' + prefix, {
                    'method': 'PUT',
                    'headers': {
                        'x-amz-acl': 'bucket-owner-read'
                    },
                    'body': file
                })
                    .then(resp => {
                        if (resp.ok) {
                            chrome.runtime.sendMessage({
                                time,
                                'title': file.name,
                                'favIconUrl': icos[file.type],
                                'prefix': prefix
                            });
                        }
                    })
                    .catch(resp => {
                        _showSnackbar(_getMessage('snackbar_error_file_load'));
                        // TODO check
                        report(
                            'Error uploading file to S3: ' + resp.url + ', status: ' + resp.status
                        );
                    });  // TODO add into i18n JSONs
            }
        };
        _gaq.push(['_trackEvent', 'ui', 'openfile']);
    };

    nodes['playPause'].onclick = function() {
        if (this.classList.contains('playing')) {
            chrome.runtime.sendMessage({'fromPopup': 'pause'});
            this.classList.remove('playing');

            const current = document.querySelector('.current');
            if (current) {
                current.classList.remove('playing');
            }
        } else {
            if (document.getElementById('list-feed').children.length) {
                // Resume current or start the 'playlist'
                (document.querySelector('.current img') || document.querySelector('.item-feed img'))
                    .click();
            } else {
                chrome.runtime.sendMessage({'fromPopup': 'resume'});
                this.classList.add('playing');
            }
        }
        this.blur();
        // Without it sometimes Spacebar push will be duplicated

        _gaq.push(['_trackEvent', 'ui', 'playpause']);
    };

    nodes['prevChunk'].onclick = function() {
        chrome.runtime.sendMessage({'fromPopup': 'prevChunk'});
        nodes['playPause'].classList.add('playing');
        _executeNowOrWhenNodesAppears(
            _ => [document.querySelector('.current')],
            node => node.classList.add('playing')
        );

        _gaq.push(['_trackEvent', 'ui', 'prev']);
    };

    nodes['nextChunk'].onclick = function() {
        chrome.runtime.sendMessage({'fromPopup': 'nextChunk'});
        nodes['prevChunk'].disabled = false;
        nodes['playPause'].classList.add('playing');
        _executeNowOrWhenNodesAppears(
            _ => [document.querySelector('.current')],
            node => node.classList.add('playing')
        );

        _gaq.push(['_trackEvent', 'ui', 'next']);
    };

    nodes['continuousPlaying'].onclick = function() {
        const val = this.classList.toggle('true');
        chrome.storage.sync.set({'isContinuousPlaying': val});
        chrome.storage.sync.get(['id_token', 'refresh_token'], tokens =>
            _keepTokenFresh(tokens).then(tokensFresh =>  // eslint-disable-line no-undef
                _setSetting('isContinuousPlaying', val, tokensFresh['id_token'])
            )
        );

        _gaq.push(['_trackEvent', 'ui', 'continuousplaying']);
    };

    nodes['listFeed'].onclick = function(mouseEvent) {
        console.log('clicked in list-feed', mouseEvent);

        if (this.classList.contains('disabled')) return;
        if (mouseEvent.target.classList.contains('addingToFeed')) {
            _gaq.push(['_trackEvent', 'ui', 'addingtofeed-item-disabled']);
            return;
        }

        if (mouseEvent.target.tagName === 'IMG') {
            // Click to the :before shows in mouseEvent.target only parent item-feed
            if (mouseEvent.target.parentElement.classList.contains('playing')) {
                chrome.runtime.sendMessage({'fromPopup': 'pause'});
                mouseEvent.target.parentElement.classList.remove('playing');
                nodes['playPause'].classList.remove('playing');

                _gaq.push(['_trackEvent', 'ui', 'item-pause']);
            } else {
                if (mouseEvent.target.parentElement.classList.contains('addingToFeed')) {
                    _gaq.push(['_trackEvent', 'ui', 'addingtofeed-item-disabled']);
                    return; // TODO instead play after added?
                }

                nodes['playPause'].classList.add('playing');

                if (mouseEvent.target.parentElement.classList.contains('current')) {
                    chrome.runtime.sendMessage({'fromPopup': 'resume'});
                    mouseEvent.target.parentElement.classList.add('playing');
                } else {
                    this.classList.add('disabled');

                    // Because maybe something alredy playing and user switch to another item
                    const current = document.getElementsByClassName('current')[0];
                    current && current.classList.remove('current', 'playing');

                    const id = mouseEvent.target.parentElement.id;
                    chrome.runtime.sendMessage(
                        {
                            'fromPopup': 'playFromFeed',
                            'id': id
                        }, () => {
                            mouseEvent.target.parentElement.classList.add('current', 'playing');
                            nodes['playPause'].disabled = true;
                            // May be current on pause
                        }
                    );
                    _gaq.push(['_trackEvent', 'ui', 'item-play']);
                }

            }
        } else if (mouseEvent.target.tagName === 'A') {
            _gaq.push(['_trackEvent', 'ui', 'item-openlink']);

            // https://developer.chrome.com/extensions/tabs#method-create
            chrome.tabs.create({'url': mouseEvent.target.innerText});
        } else if (mouseEvent.target.classList.contains('ico-download')) {
            if (isNaN(
                mouseEvent.target.nextSibling.innerText[
                    mouseEvent.target.nextSibling.innerText.length - 1
                ]
            )) {
                chrome.runtime.sendMessage({
                    'fromPopup': 'downloadLocalOGGFromIndexedDB',
                    'id': mouseEvent.target.parentElement.id
                });
            }
            _gaq.push(['_trackEvent', 'ui', 'item-download']);
        } else if (mouseEvent.target.classList.contains('ico-delete')) {
            _showDialog(
                mouseEvent.target.parentElement.querySelector('a').innerText,
                _ => {
                    _closeDialog();
                    _progressShow();
                    removeFeedItem(+mouseEvent.target.parentElement.id);
                },
                'button_cancel',
                'delete'
            );
            _gaq.push(['_trackEvent', 'ui', 'item-delete']);
        } else {
            if (mouseEvent.target.parentElement.id.match(/\d+/)) {
                _togglePlayedStatus(mouseEvent.target.parentElement);
                _gaq.push(['_trackEvent', 'ui', 'item-toggleplayedstatus']);
            }
        }
    };

    nodes['playbackRates'].forEach(node => node.onchange = function() {
        const playbackRate = +this.value;
        chrome.runtime.sendMessage({
            'fromPopup': 'alterPlaybackRate',
            'value': playbackRate
        });
        chrome.storage.sync.set({playbackRate});
        _keepTokenFresh(storageCache).then(tokensFresh =>  // eslint-disable-line no-undef
            _setSetting('playbackRate', playbackRate, tokensFresh['id_token'])
        );

        _gaq.push(['_trackEvent', 'ui', 'playbackrate']);
    });

    document.onkeydown = event => {

        if (event.code === 'Space' && document.activeElement.id !== 'search') {
            event.preventDefault(); // Do not scroll the list
            nodes['playPause'].click();

            _gaq.push(['_trackEvent', 'ui', 'playpause-hotkey']);
        }

    };

    console.timeEnd('initEventListeners');
}


function _showDialog(text, callback, ...buttonsTexts) {
    // TODO Spanish where '¿' before word too
    nodes['dialog'].querySelector('header h2').innerText = _getMessage('delete') + '?';

    nodes['dialog'].querySelector('section').innerText = text;
    nodes['dialog'].querySelectorAll('button').forEach(
        (btn, i) => {
            btn.innerText = _getMessage(buttonsTexts[i]);
            btn.onclick = i ? callback : _closeDialog;
        }
    );

    nodes['dialog'].classList.add('mdc-dialog--open');
}
function _closeDialog() {
    nodes['dialog'].classList.remove('mdc-dialog--open');
}
async function removeFeedItem(id) {
    const {item, isUniqueForUser} = await _prepareRemoving(id);
    console.debug('Item for removing:', item);

    /* We send UTC instead of pubDate from `new Date(Number(1000) * 1000).toUTCString()`
     * because Chrome and Firefox retunrs different formats,
     * and maybe this will be changed again.
     * Now - less code, less paydata, but more cycles on Lambda.
     */
    _keepTokenFresh(storageCache).then(_storageCache =>  // eslint-disable-line no-undef
        fetch(stageUrlNew + 'articles/' + item['hashArticle'] +  // eslint-disable-line no-undef
                '?voice='+item['voice']+
                '&lenChunks='+item['lenChunks']+
                '&utc='+id+
                '&isUniqueForUser='+isUniqueForUser, {
            'method': 'DELETE',
            'headers': {
                'Authorization': storageCache['id_token'],
            }
        })
            .then(_ => {
                _deleteArticle(id);  // eslint-disable-line no-undef
                document.getElementById(id).remove();
                if (!document.getElementById('list-feed').children.length) {
                    nodes['playPause'].disabled = true;
                }
                _showSnackbar(_getMessage('snackbar_itemRemoved'));
            }).catch(_ => {
                _showSnackbar(_getMessage('snackbar_error_feedItemRemoving'));
            }).finally(_ => {
                _progressHide();
            })
    );
    if (document.getElementById(id).classList.contains('playing')) {
        chrome.runtime.sendMessage({'fromPopup': 'stopPlaying'});
    }
}
function _prepareRemoving(id) {
    console.debug('Removing - get item - and flag is unique, id:', id);
    return new Promise(resolve => {
        indexedDB.open('intelligent-speaker').onsuccess = event => {
            try {
                event.target.result
                    .transaction('articles')
                    .objectStore('articles')
                    .get(id)
                    .onsuccess = event => {
                        const item = event.target.result;
                        if (!item) {
                            /** This is the debug case, added because of many emails with errors:
                            /* "Cannot read property 'hashArticle' of undefined".
                             * TODO in such case remove the item locally only?
                             */
                            throw 'No item for id:' + id;
                        }
                        console.debug('Searching for items with the same hash, current:', item);
                        event.target.source.openCursor()
                            .onsuccess = event => {
                                const cursor = event.target.result;
                                if (cursor) {
                                    if (cursor.value['hashArticle'] === item['hashArticle'] &&
                                        cursor.key !== id)
                                        resolve({item, 'isUniqueForUser': false});
                                    cursor.continue();
                                }
                                else {
                                    resolve({item, 'isUniqueForUser': true});
                                }
                            };
                    };
            } catch(ex) {
                report('ERROR _prepareRemoving:', ex, ', id:', id);
            }
        };
    });
}


/**
 * Not storageCache becaue may be the case when after first tts without closing of the UI
 * (without reiniting of the storageCache) here we have storageCache without id_token.
 *
 * Closing of the popup during fetch do not abort http call, I tried.
 */
function emailPodcastLink(storageSync) {
    fetch(`${stageUrl}send-email-podcast`, {  // eslint-disable-line no-undef
        'headers': {
            'Authorization': storageSync['id_token'],
            'Content-Type': 'application/json'  // for body mapping template in ApiGateway
        }
    }).then(_ => _showSnackbar(_getMessage('snackbar_podcastEmailSent')))
        .catch(_ => _showSnackbar(_getMessage('snackbar_error_podcastEmail')))
        .then(_ => {
            _progressHide();
        });
}


function _progressShow() {
    nodes['progress'] = document.getElementById(
        'progress--this-is-some-random-id-name-that-do-not-collide-with-existing-id-in-dom'),

    nodes['progress'].style.display = 'block'; // Because was hidden by _setUIBeforeLogin()
    nodes['progress'].firstElementChild.style.display = 'block';
    nodes['progress'].classList.add(
        'intelligentspeaker--active--please-do-not-collide');
}
function _progressHide() {
    nodes['progress'].classList.remove(
        'intelligentspeaker--active--please-do-not-collide');
}

function _showSnackbar(text) {
    nodes['snackbar'].innerText = text;
    nodes['snackbar'].classList.add(
        'intelligentspeaker--snackbar--class_active');
    setTimeout(
        _ => nodes['snackbar'].classList.remove(
            'intelligentspeaker--snackbar--class_active'),
        5000
    );
}


function _fillListAudio() {

    const fillPart = articles => {
        const _listFeedChildren = [];
        console.debug('Number of articles from IndexedDB:', articles.size);

        const callback = _ => {
            if (!nodes['listFeed']) {
                nodes['listFeed'] = document.getElementById('list-feed');
            }
            console.debug(_listFeedChildren);

            console.time('append');
            nodes['listFeed'].append(..._listFeedChildren);
            console.timeEnd('append');
        };

        articles.forEach((article, id) => {
            const cacheStatusForUser = _checkCache(id, article);
            _listFeedChildren.push(_buildFeedEntity(
                id,
                article['url'],
                article['title'],
                article['favIconUrl'],
                false,
                false,
                article['isNew'],
                cacheStatusForUser,
                article['isAddedOffline']));
        });
        if (['interactive', 'complete'].includes(document.readyState)) {
            console.log('Filling feed items: DOM was ready without listeners');
            callback();
        } else {
            document.onreadystatechange =
                event => {
                    if (event.target.readyState === 'interactive') {
                        console.log('Filling feed items: from DOM event');
                        callback();
                    }
                };
        }
    };

    const amountOfVisibleItems = 8;
    _getArticles(0, amountOfVisibleItems).then(articles => {
        if (articles.size) {
            // Reset UI on init history sync (after login), against duplications of articles
            if (!nodes['listFeed']) {
                nodes['listFeed'] = document.getElementById('list-feed');
            }
            nodes['listFeed'].classList.remove('empty');
            nodes['listFeed'].innerHTML = '';
        } else {
            document.getElementById('list-feed').classList.add('empty');
            document.getElementById('play-pause').disabled = true;
            return;
        }

        fillPart(articles);

        // Not nodes.playPause because we fill list before nodes
        document.getElementById('play-pause').disabled = false;

        _getArticles(amountOfVisibleItems, 9999999).then(articles => {
            if (articles) {
                fillPart(articles);
            }
        });
    });
}
function _checkCache(id, article) {
    if (article['cache'] && article['lenChunks'] === article['cache'].length) {
        return;
    }

    console.debug(
        'Now will fetch missing cache for this title:', article['title'],
        'url:'                                        , article['url'] );

    // Postpone for faster UI init
    setTimeout(_ => {
        chrome.runtime.sendMessage({
            'fromPopup': '_cacheExisting',
            id,
            'hashArticle': article['hashArticle'],
            'lenChunks': article['lenChunks'],
            'idToken': storageCache['id_token']
        });
    }, 1000);

    return ` ${article['cache'] ? article['cache'].length : 1} / ${article['lenChunks']}`;
}


/** We save to browser storage and backend ico as link -
 *  neither as blob nor base64 becasue due to the CORS
 *  we can not fetch data of ico from many websites
 *  (neither using fetch nor img.src with reading of data from canvas);
 *  maybe in future if user will ask us we will save base64 of icons from some websites -
 *  for case when user add to the feed from pc in intranet and
 *  view feed from home without vpn.
 *
 *  Name of the property is 'isNew' but not 'new' because in JS
 *  we can not create value with such name.
 *
 *  TODO encrypt data in IndexedDB?
 */
function _persistItemFeed(
    time,
    favIconUrl,
    url,
    title,
    isAddingToFeed=false,
    isNew=false,
    isPlaying=false,
    hashArticle,
    lenChunks) {

    nodes['listFeed'].prepend(_buildFeedEntity(
        time,
        url,
        title,
        favIconUrl,
        isPlaying,
        isAddingToFeed,
        isNew));

    /**
     *  Do not JSON.parse(JSON.stringify... because it turns lenChunks into String
     *  and from get-presigned-urls we will get exception `
     *  'str' object cannot be interpreted as an integer
     *  `
     */
    const feedEntityNew = {
        favIconUrl,
        url,
        title,
        isNew,
        lenChunks
    };

    /** Set only when this is from history sync and we have full data,
     * will not saved adding to the feed */
    if (hashArticle) {
        feedEntityNew['hashArticle'] = hashArticle;
        feedEntityNew['lenChunks'] = lenChunks;
        _insertOrReplace(time, feedEntityNew);  // eslint-disable-line no-undef
    }
}


/**
 *  Not hand-written html with assign into innerHTML
 *  because reviewers from stores do not like it,
 *  their lintings warn such 'eval' style.
 */
function _buildFeedEntity(
    time,
    url,
    title,
    favIconUrl='images/logo/logo-with-bigger-margin.png',
    isPlaying=false,
    isAddingToFeed=false,
    isNew=false,
    cacheStatusForUser='',
    isAddedOffline=false) {

    const li = document.createElement('li');
    li.id = time;
    li.classList.add('item-feed');
    if (isPlaying) {
        li.classList.add('current', 'playing');
    }
    if (isAddingToFeed) {
        li.classList.add('addingToFeed', 'new');
    }
    if (isNew) {
        li.classList.add('new');
    }
    if (isAddedOffline) {
        li.classList.add('isAddedOffline');
    }

    const img = document.createElement('img');
    img.classList.add('favIcon');
    if (favIconUrl) {
        img.src = favIconUrl;
    }
    li.appendChild(img);

    const wrapper = document.createElement('div');
    wrapper.classList.add('wrapper--url-and-title');

    const a = document.createElement('a');
    a.textContent = url;
    wrapper.appendChild(a);

    const titleNode = document.createElement('span');
    titleNode.classList.add('title');

    /**
     *  Here we replace '<' in title for '&lt;' - this is enough for
     *  showing html tags, for example when title of the page
     *  'This is the article about history of <ol> tag'.
     *  Without this replace will show
     *  'This is the article about history of tag'.
     *  '
     */
    titleNode.textContent = title.replace('<', '&lt;');
    wrapper.appendChild(titleNode);

    li.appendChild(wrapper);

    const downloadButton = document.createElement('span');
    downloadButton.textContent = 'save_alt';
    downloadButton.classList.add('ico-download');
    li.appendChild(downloadButton);
    const tooltipDownload = document.createElement('span');
    tooltipDownload.classList.add('tooltip');
    li.appendChild(tooltipDownload);

    const deleteButton = document.createElement('span');
    deleteButton.textContent = 'delete';
    deleteButton.classList.add('ico-delete');
    li.appendChild(deleteButton);
    const tooltipDelete = document.createElement('span');
    tooltipDelete.classList.add('tooltip');
    li.appendChild(tooltipDelete);

    const callbackSetTooltipTexts = _ => {
        if (isAddingToFeed) {
            tooltipDownload.textContent = _getMessage('tooltip_download_caching')
                + cacheStatusForUser;
        } else {
            tooltipDownload.textContent = _getMessage('tooltip_download');
        }

        tooltipDelete.textContent = _getMessage('delete');
    };

    if (messages) {
        callbackSetTooltipTexts();
    } else {
        addEventListener('messagesIsReady', callbackSetTooltipTexts);
    }

    return li;
}


/**
 *  Maybe user added article on another browser
 *  or toggled played status for a few items.
 */
async function _syncHistory(updated=0, id_token) {
    const resp = await fetch(`${stageUrlNew}history?updated=${updated}`, {
        'headers': {
            'Authorization': id_token
        }
    });
    const resp_json = await resp.json();
    console.debug('_syncHistory(): before waiting all promises');

    if (!resp_json['Items']) {
        throw '_syncHistory: incorrect response from the backend?, resp_json:' +
            JSON.stringify(resp_json);
    }

    await Promise.all(resp_json['Items'].map(item => _syncHistoryItem(item, id_token)));
    console.debug('_syncHistory(): after resolving all promises');
    chrome.storage.sync.set({'updated': Date.now() / 1000 | 0});

}async function _syncHistoryItem(item, id_token) {
    console.log('Updated history item from the backend:', item);
    const id = +[item['utc']['N']];
    if (!item['is_deleted'] && !await getFromObjectStore(id)) {
        // Item was added not from this client (another browser, email) - lets ingest it.
        /** Just FYI: you can transform response from DynamoDB in API Gateway.
         *  Cons: exported json of API Gateway (dump)
         *  will show your Apache Velocity code in one line.
         */
        _persistItemFeed(
            id,

            item['favIconUrl'] ? item['favIconUrl']['S'] : '',
            /* Some pages without ico, for example
             * http://reports.weforum.org/future-of-jobs-2016/skills-stability
             * Keys with undefined values will automatically dropped from Chrome Storage. */

            // No url_article for added files
            item['url_article'] ? item['url_article']['S'] : '',
            item['title']['S'],
            false,

            item['isNew'] && item['isNew']['BOOL'],
            /* We check existance of item['isNew'] because
             * history from old users without 'isNew'.
             */

            false,
            item['hash_article']['S'],

            +item['len_chunks']['N']
            /* Number() against error in lambda get-presigned-urls: `
             * 'str' object cannot be interpreted as an integer` */
        );
        chrome.runtime.sendMessage({
            'fromPopup': '_cacheExisting',
            'id': item['utc']['N'],
            'hashArticle': item['hash_article']['S'],
            'lenChunks': item['len_chunks']['N'],
            'idToken': id_token
        });
        nodes['listFeed'].classList.remove('empty');  // For the first sync
    } else if (!item['is_deleted']) {
        console.log('Lets update played status from backend');
        _updateIDB(
            +item['utc']['N'],
            'isNew',
            item['isNew']['BOOL']
        );

        _executeNowOrWhenNodesAppears(
            _ => [document.getElementById(item['utc']['N'])],
            itemFeed => itemFeed.classList.toggle('new', item['isNew']['BOOL'])
        );

    } else if (item['is_deleted']) {
        console.log('From backend: item deleted');
        _deleteArticle(id);  // eslint-disable-line no-undef
        const nodeForDeleting = document.getElementById(id);

        // IF because sometimes signal from DB that should be deleted but already not exists
        if (nodeForDeleting) {
            nodeForDeleting.remove();
        }
    }
}


function _updateUser(id_token, isUserWithConfirmedEmail=false) {
    fetch(`${stageUrlNew}user`, {  // eslint-disable-line no-undef
        'headers': {
            'Authorization': id_token
        }
    })
        .then(resp => resp.json())
        .then(resp => {
            console.log('User table:', resp);

            const o = {};

            o['chars_left'] = resp['CharsLeft'];

            if (isUserWithConfirmedEmail) {

                if (resp['IsContinuousPlaying']) {
                    o['isContinuousPlaying'] = resp['IsContinuousPlaying'];
                }
                if (resp['PlaybackRate']) {
                    o['playbackRate'] = resp['PlaybackRate'];
                }

                if (resp['Voices']) {
                    o['voices'] = resp['Voices'];
                }

                if ('IsSubscribed' in resp) {
                    o['isSubscribed'] = resp['IsSubscribed'];
                }
            }

            chrome.storage.sync.set(o);
        });
}

function _minutesToHumanText(minutes) {
    return `${minutes / 60 | 0}h : ${minutes % 60 | 0}m`;
}


/**
 *  We use this util function instead of pure Date.now() / 1000 | 0
 *  because user can add to the feed and play current page
 *  at the same second.
 *
 *  We use unixtime in seconds instead of milliseconds
 *  for smaller size of the database.
 *
 *  One potential pitfall about this -
 *  user can play at the same second
 *  from multiple clients.
 */
function _getNextUTCId() {
    // The last on the top
    const nodeLastItemFeed = document.querySelector('.item-feed');
    const unixtime = Date.now() / 1000 | 0;

    // Check existance of node because after install will be zero items
    if (nodeLastItemFeed && nodeLastItemFeed.id == unixtime)
        return unixtime + 1;
    return unixtime;
}


function _togglePlayedStatus(node) {
    const val = node.classList.toggle('new');
    chrome.runtime.sendMessage({
        'fromPopup': 'togglePlayedStatus',

        'id_token': storageCache['id_token'],
        'refresh_token': storageCache['refresh_token'],

        [node.id]: storageCache[node.id],

        'id': node.id,
        'val': val
    });
}


async function _setSetting(name, val, id_token) {
    // TODO drop ApiGateway - POST directly to the DynamoDB - will be faster, cheaper, simpler.
    fetch(`${stageUrlNew}setting`, {  // eslint-disable-line no-undef
        'method': 'PUT',
        'headers': {
            // eslint-disable-next-line no-undef
            'Authorization': id_token || (await _keepTokenFresh())['id_token'],
            'Content-Type': 'application/json'
        },
        'body': JSON.stringify({[name]: val})
        // Without stringify browser will not send the body
    });
}

function _openSitePage(pageName='') {
    chrome.tabs.create({
        'url': `https://intelligent-speaker.com/${pageName}/` +
        '?id_token=' + storageCache['id_token'] +
        '&refresh_token=' + storageCache['refresh_token']
    });
}
